/* Element
   #Mm2238
   Lutz Prechelt
   1997-01-02
   RCS: $Id: ElementMain.java,v 1.2 1997/01/22 16:42:38 prechelt Exp prechelt $
*/
import java.util.Enumeration;
import java.util.Vector;

/*
Dieses Programm verwaltet Und/Oder-Sequenzen (And/Or-Sequenzen) von Strings.
Solche Sequenzen bestehen aus 'Element'-Objekten. Es gibt drei Arten davon:
- StringElement enthaelt einen einzelnen String.
- AndElement enthaelt eine Liste von Elementen, die in der gegebenen 
  Reihenfolge verkettet sind.
- OrElement enthaelt eine Menge von Elementen, von denen je
  eines alternativ eingesetzt benutzt kann.
AndElement- und OrElement-Objekte koennen beliebig verschachtelt werden.
Die unterste Ebene (Blaetter) eines dabei entstehenden Objektbaums wird immer
von StringElement-Objekten gebildet.
OrElemente fuehren zu Varianten. Treffen mehrere OrElemente zusammen,
wird das Kreuzprodukt ihrer Varianten gebildet.

Hier die Klassenhierarchie:
  abstract class Element
     class StringElement
     class AndElement
     class OrElement
  interface ElementAction
     class Depth
  class ElementMain

Die letzte Klasse enthaelt ein Hauptprogramm, das eine And/Or-Sequenz
erzeugt und ausgibt.
#M
#M*** ENTWURFSMUSTER: ***
#M
#MElement ist die abstrakte Oberklasse eines **Kompositum**.
#MStringElement ist der zugehoerige Einzelteiltyp (Blatt),
#MAndElement und OrElement sind zugehoerige Behaeltertypen (Kompositum).
#M
#MElementAction ist die abstrakte Oberklasse eines **Besuchers**.
#MDie besuchte Datenstruktur sind verschachtelte Element-Behaelter.
*/


//------------------------ Element ------------------------------
/**
  Schnittstelle der 'Element'-Klassen StringElement, AndElement, OrElement.
#M  *** ENTWURFSMUSTER: ***
#M  Element ist die Oberklasse in einem Kompositum-Muster.
*/
abstract class Element {
  /** Zufuegeoperation. Fuegt, falls moeglich, diesem Element ein weiteres
      Element zu.
  */
  abstract public void add(Element e);
  
  /** Repraesentationsoperation. Liefert eine kodierte Darstellung des
      Elements als String.
  */
  abstract public String asString();
  
  /** Ausgabeoperation. Gibt das gesamte Element auf System.out aus.
      Jede Variante erscheint auf einer eigenen Zeile.
  */
  public void print() {
     Vector v = variants();
     Enumeration e = v.elements();
     while (e.hasMoreElements())
        System.out.println((String)e.nextElement());
  }
  
  /** Repraesentationsoperation. Liefert einen Vektor, in dem jedes Element
      die Stringrepraesentation einer Variante der And/Or-Sequenz enthaelt.
      Jede Variante ist genau einmal im Vektor vertreten.
  */
  abstract public Vector variants();
  
  /** Verzweigungsoperation. Ruft die zur jeweiligen Unterklasse gehoerige
      Operation von ElementAction auf.
#M      *** ENTWURFSMUSTER: ***
#M      Dies ist die 'double dispatch' Prozedur in der besuchten Datenstruktur
#M      fuer das Besucher-Muster.
  */
  abstract void perform(ElementAction a);
}


//------------------------ StringElement ------------------------------
/**
  Elementklasse, die genau einen String enthaelt.
#M  *** ENTWURFSMUSTER: ***
#M  StringElement ist die (einzige) Blattklasse in einem Kompositum-Muster.
*/
class StringElement extends Element {
  String inhalt;     // Inhalt des StringElements

  /** Erzeugt ein StringElement mit leerem Inhalt.
  */
  public StringElement() {
     inhalt = new String();
  }

  /** Erzeugt ein StringElement mit dem als Argument gegebenen Inhalt.
  */
  public StringElement(String s) {
     inhalt = new String(s);
  }

  /** Zufuegeoperation. Fuegt e an bisherigen Stringinhalt an,
      falls e ein Stringelement ist.
      Wirft andernfalls CastException.
  */
  public void add(Element e) {
     StringElement s = (StringElement)e; // wirft evtl. CastException
     inhalt = inhalt + s.inhalt;
  }

  /** Repraesentationsoperation. Liefert eine kodierte Darstellung des
      Elements als String in der Form
      "inhalt"
  */
  public String asString() {
     return ("\"" + inhalt + "\"");
  }
  
  /** Liefert einen Vektor mit genau einem Element, naemlich dem String,
      den das StringElement enthaelt.
  */
  public Vector variants() {
     Vector v = new Vector();
     v.addElement((Object)inhalt);
     return v;
  }
  
  /** Ruft die zu der uebergebenen ElementAction gehoerende stringAction auf.
#M      *** ENTWURFSMUSTER: ***
#M      perform() ist die Verzweigungsoperation fuer Besucher (ElementAction).
  */
  void perform(ElementAction a) {
     a.stringAction(this);
  }
}


//------------------------ AndElement ------------------------------
/**
  Und-Sequenz von Elementen.
  (Die Elemente sind also miteinander zu verketten.)
#M  *** ENTWURFSMUSTER: ***
#M  AndElement ist eine der Behaelterklassen in einem Kompositum-Muster.
*/
class AndElement extends Element {
  Vector elems;   // Vektor der Elemente, die zusammen das AndElement bilden

  /** Erzeugt ein AndElement ohne Inhalt.
  */
  public AndElement() {
     elems = new Vector();
  }

  /** Zufuegeoperation. Fuegt e als neues Element an bisherige Sequenz von
      Elementen an.
  */
  public void add(Element e) {
     elems.addElement(e);
  }

  /** Repraesentationsoperation. Liefert eine kodierte Darstellung des
      Elements als String in der Form
      UND(el1&el2&el3)
  */
  public String asString() {
     StringBuffer b = new StringBuffer("UND(");
     Enumeration n = elems.elements();
     while (n.hasMoreElements()) {
        Element e = (Element)n.nextElement();
        b.append(e.asString());
        b.append(n.hasMoreElements() ? "&" : ")");
     }
     return b.toString();
  }
  
  /** Liefert das Kreuzprodukt saemtlicher Varianten saemtlicher Elemente
      des AndElements. Die Produktbildung geschieht durch Stringverkettung der
      Varianten.
  */
  public Vector variants() {
     Vector result = new Vector();
     Enumeration n = elems.elements();
     while (n.hasMoreElements()) {
        Element e = (Element)n.nextElement();
        Vector nextpart = e.variants();
        if (nextpart.size() == 0)
           continue;  // nothing to add to result for this element
        if (result.size() == 0) {
           result = nextpart;
           continue;  // nothing to combine with yet
        }
        Vector oldresult = result;
        int oldN  = oldresult.size();
        int nextN = nextpart.size();
        result = new Vector(oldN * nextN);
        // erzeuge Kreuzprodukt (in Form von Stringverkettung):
        // result := oldresult x nextpart
        for (int oldI = 0; oldI < oldN; oldI++) {
           String currentold = (String)oldresult.elementAt(oldI);
           for (int nextI = 0; nextI < nextN; nextI++) {
              String currentnext = (String)nextpart.elementAt(nextI);
              result.addElement(currentold + currentnext); // Stringverkettung
           }
        }
     }
     return result;
  }
  
  /** Ruft die zu der uebergebenen ElementAction gehoerende andAction auf.
#M      *** ENTWURFSMUSTER: ***
#M      perform() ist die Verzweigungsoperation fuer Besucher (ElementAction).
  */
  void perform(ElementAction a) {
     a.andAction(this);
  }
}


//------------------------ OrElement ------------------------------
/**
  Oder-Menge von Elementen.
  (Die Elemente sind also alternativ.)
#M  *** ENTWURFSMUSTER: ***
#M  OrElement ist eine der Behaelterklassen in einem Kompositum-Muster.
*/
class OrElement extends Element {
  Vector elems;   // Vektor der Elemente, die zusammen das OderElement bilden

  /** Erzeugt ein OderElement ohne Inhalt.
  */
  public OrElement() {
     elems = new Vector();
  }

  /** Zufuegeoperation. Fuegt e als neues Element in bisherige Menge von
      alternativen Elementen ein.
  */
  public void add(Element e) {
     elems.addElement(e);
  }

  /** Repraesentationsoperation. Liefert eine kodierte Darstellung des
      Elements als String in der Form
      ODER(el1|el2|el3)
  */
  public String asString() {
     StringBuffer b = new StringBuffer("ODER(");
     Enumeration n = elems.elements();
     while (n.hasMoreElements()) {
        Element e = (Element)n.nextElement();
        b.append(e.asString());
        b.append(n.hasMoreElements() ? "|" : ")");
     }
     return b.toString();
  }
  
  /** Liefert die Vereinigungsmenge aller Varianten aller Elemente
      des OderElements.
  */
  public Vector variants() {
     Vector result = new Vector();
     Enumeration n = elems.elements();
     while (n.hasMoreElements()) {
        Element e = (Element)n.nextElement();
        Vector nextpart = e.variants();
        Enumeration nn = nextpart.elements();
        result.ensureCapacity(result.size() + nextpart.size());
        while (nn.hasMoreElements())
           result.addElement(nn.nextElement());
     }
     return result;
  }
  
  /** Ruft die zu der uebergebenen ElementAction gehoerende orAction auf.
#M      *** ENTWURFSMUSTER: ***
#M      perform() ist die Verzweigungsoperation fuer Besucher (ElementAction).
  */
  void perform(ElementAction a) {
     a.orAction(this);
  }
}


//------------------------ ElementAction ------------------------------
/**
  Schnittstelle fuer Operationen auf ElementStrukturen, bei denen die
  einzelnen Teile je nach ihrem Typ unterschiedlich behandelt werden sollen.
#M  *** ENTWURFSMUSTER: ***
#M  Entwurfsmuster 'Besucher' (Visitor): ElementAction ist die Oberklasse aller 
#M  Besucher fuer 'Element'-Kompositumdatenstrukturen.
*/
interface ElementAction {
  void stringAction(StringElement e);
  void andAction(AndElement e);
  void orAction(OrElement e);
}


//------------------------- Depth ------------------------------
/**
  Klasse zum Berechnen von Maximaltiefen der verschiedenen Knotenarten in
  einem And/Or-Elementbaum.
#M  *** ENTWURFSMUSTER: ***
#M  Depth ist ein Besucher fuer die Kompositum-Datenstruktur Element.
#M  Element.perform() ist die Verzweigungsoperation fuer den Besucher.
*/
class Depth implements ElementAction {
  public  int maxDepth      = -1;  // Tiefe des tiefsten Elements
  public  int maxAndDepth   = -1;  // Tiefe des tiefsten AndElements
  public  int maxOrDepth    = -1;  // Tiefe des tiefsten OrElements
  private int depth         =  0;  // aktuelle Tiefe (Wurzel == 0)
  
  /** Durchwandert das Element e, um sich zu initialisieren.
      Die Komponenten maxDepth, maxAndDepth und maxOrDepth koennen 
      anschliessend abgefragt werden.
  */
  public Depth(Element e) {
     e.perform(this);
  }
  
  public void stringAction(StringElement e) {
     if (maxDepth < depth)
        maxDepth = depth;
  }
  
  public void andAction(AndElement u) {
     if (maxAndDepth < depth) {
        maxAndDepth = depth;
        if (maxDepth < depth)
           maxDepth = depth;
     }
     iterate(u.elems.elements());
  }
  
  
  public void orAction(OrElement o) {
     if (maxOrDepth < depth) {
        maxOrDepth = depth;
        if (maxDepth < depth)
           maxDepth = depth;
     }
     iterate(o.elems.elements());
  }
  
  /** Durchlaufe die enthaltenen Elemente eines AndElement oder OrElement.
  */
  private void iterate(Enumeration n) {
     depth++;  // enthaltene Elemente liegen eine Ebene tiefer.
     while (n.hasMoreElements()) {
        Element e = (Element)n.nextElement();
        e.perform(this);
     }
     depth--;
  }
}


//------------------------ ElementMain ------------------------------
/**
  Hauptprogramm. Erzeugt eine And/Or-Sequenz und gibt sie aus.
*/
public class ElementMain {
  public static void main (String args[]) {
     AndElement u = new AndElement();
     OrElement modal = new OrElement();
     OrElement verb  = new OrElement();
     u.add(new StringElement("Wer "));
     modal.add(new StringElement("kann "));
     modal.add(new StringElement("will "));
     u.add(modal);
     u.add(new StringElement("denn schon "));
     verb.add(new StringElement("denken "));
     verb.add(new StringElement("handeln "));
     u.add(verb);
     u.add(new StringElement("?"));

     u.print();
     System.out.println(u.asString());
     Depth t = new Depth(u);
     System.out.println("  Tiefe=" + t.maxDepth +
                        "  UndTiefe=" + t.maxAndDepth +
                        "  OderTiefe=" + t.maxOrDepth);
  }
}
