/* Element
   #Mm2238
   Lutz Prechelt
   1997-01-02
   RCS: $Id: ElementMain.java,v 1.2 1997/01/22 16:42:38 prechelt Exp prechelt $
*/
import java.util.Enumeration;
import java.util.Vector;

/*
This program manages And/Or-Sequences of Strings.
Such sequences consist of 'Element' objects. There are three kinds of these:
- StringElement contains a single String.
- AndElement contains a list of elements that are concatenated in the given
  order.
- OrElement contains a set of elements that are used alternatively,
  exactly one at a time.
AndElement- and OrElement objects can be nested arbitrarily.
The lowest level (leafs) of any thus created object tree is always formed
by StringElement objects.
OrElements result in variants. Where OrElements meet, the cross product
of their variants is formed.

Here is the class hierarchy:
  abstract class Element
     class StringElement
     class AndElement
     class OrElement
  interface ElementAction
     class Depth
  class ElementMain

ElementMain contains a main program that generates an And/Or sequence
and prints it.
#M
#M*** DESIGN PATTERN: ***
#M
#MElement is the abstract superclass of a **Composite**.
#MStringElement is its leaf type.
#MAndElement and OrElement are its container (composite) types
#M
#MElementAction is the abstract superclass of a **Visitor**.
#MThe data structure that is visited is an Element container nesting.
*/


//------------------------ Element ------------------------------
/**
  Interface of the 'Element' classes StringElement, AndElement, OrElement.
#M  *** DESIGN PATTERN: ***
#M  Element is the superclass of a Composite pattern.
*/
abstract class Element {
  /** include operation. Adds another element to this element if possible.
  */
  abstract public void add(Element e);
  
  /** repraesentation operation. Returns a coded representation of the
      Element as a String.
  */
  abstract public String asString();
  
  /** output operation. Prints the whole Element on System.out
      Each variant is printed on a line of its own.
  */
  public void print() {
     Vector v = variants();
     Enumeration e = v.elements();
     while (e.hasMoreElements())
        System.out.println((String)e.nextElement());
  }
  
  /** repraesentation operation. Returns a vector in which each element
      contains the String representation of one variant of the
      And/Or sequence.
      Each variant appears exactly once in the vector
  */
  abstract public Vector variants();
  
  /** Branching operation. Calls that one operation from ElementAction that
      is responsible for (corresponds to) the present subclass.
#M      *** DESIGN PATTERN: ***
#M      This is the 'double dispatch' procedure in the visited data
#M      structure for the Visitor pattern.
  */
  abstract void perform(ElementAction a);
}


//------------------------ StringElement ------------------------------
/**
  Element class that contains exactly one String.
#M  *** DESIGN PATTERN: ***
#M  StringElement is the (only) leaf class in a Composite pattern.
*/
class StringElement extends Element {
  String contents;     // Contents of the StringElement

  /** Generates a StringElement with empty contents.
  */
  public StringElement() {
     contents = new String();
  }

  /** Generates a StringElement with the contents given as argument.
  */
  public StringElement(String s) {
     contents = new String(s);
  }

  /** include operation. Appends e to current String contents
      if e is a Stringelement.
      Throws CastException otherwise.
  */
  public void add(Element e) {
     StringElement s = (StringElement)e; // may throw CastException
     contents = contents + s.contents;
  }

  /** repraesentation operation. Returns a coded representation of
      the element as a string of the form
      "contents"
  */
  public String asString() {
     return ("\"" + contents + "\"");
  }
  
  /** Returns a vector with exactly one element: the String that the
      StringElement contains.
  */
  public Vector variants() {
     Vector v = new Vector();
     v.addElement((Object)contents);
     return v;
  }
  
  /** Calls the stringAction of the ElementAction given as argument.
#M      *** DESIGN PATTERN: ***
#M      perform() is the dispatch operation for Visitor (ElementAction).
  */
  void perform(ElementAction a) {
     a.stringAction(this);
  }
}


//------------------------ AndElement ------------------------------
/**
  And sequence of elements.
  (That means the elements have to be concatenated.)
#M  *** DESIGN PATTERN: ***
#M  AndElement is one of the container classes of a Composite pattern.
*/
class AndElement extends Element {
  Vector elems;   // vector of the elements forming the AndElement

  /** Generates an AndElement without Contents.
  */
  public AndElement() {
     elems = new Vector();
  }

  /** include operation. Adds e as new element to current sequence of
      elements.
  */
  public void add(Element e) {
     elems.addElement(e);
  }

  /** representation operation. Returns a coded representation of the
      Element as a String of the form
      AND(el1&el2&el3)
  */
  public String asString() {
     StringBuffer b = new StringBuffer("AND(");
     Enumeration n = elems.elements();
     while (n.hasMoreElements()) {
        Element e = (Element)n.nextElement();
        b.append(e.asString());
        b.append(n.hasMoreElements() ? "&" : ")");
     }
     return b.toString();
  }
  
  /** Returns cross product of all variants of all elements in the
      AndElement. The product is formed by concatenating the Strings
      of the variants.
  */
  public Vector variants() {
     Vector result = new Vector();
     Enumeration n = elems.elements();
     while (n.hasMoreElements()) {
        Element e = (Element)n.nextElement();
        Vector nextpart = e.variants();
        if (nextpart.size() == 0)
           continue;  // nothing to add to result for this element
        if (result.size() == 0) {
           result = nextpart;
           continue;  // nothing to combine with yet
        }
        Vector oldresult = result;
        int oldN  = oldresult.size();
        int nextN = nextpart.size();
        result = new Vector(oldN * nextN);
        // generate cross product (by String concatenation):
        // result := oldresult x nextpart
        for (int oldI = 0; oldI < oldN; oldI++) {
           String currentold = (String)oldresult.elementAt(oldI);
           for (int nextI = 0; nextI < nextN; nextI++) {
              String currentnext = (String)nextpart.elementAt(nextI);
              result.addElement(currentold + currentnext); // String concat
           }
        }
     }
     return result;
  }
  
  /** Calls the stringAction of the ElementAction given as argument.
#M      *** DESIGN PATTERN: ***
#M      perform() is the dispatch operation for Visitor (ElementAction).
  */
  void perform(ElementAction a) {
     a.andAction(this);
  }
}


//------------------------ OrElement ------------------------------
/**
  Or set of elements.
  (That means the elements are alternatives.)
#M  *** DESIGN PATTERN: ***
#M  OrElement is one of the container classes of a Composite pattern.
*/
class OrElement extends Element {
  Vector elems;   // vector of the elements forming the OrElement

  /** Generates an OrElement without Contents.
  */
  public OrElement() {
     elems = new Vector();
  }

  /** include operation. Adds e as new element to current set of
      alternative elements.
  */
  public void add(Element e) {
     elems.addElement(e);
  }

  /** representation operation. Returns a coded representation of the
      Element as a String of the form
      OR(el1|el2|el3)
  */
  public String asString() {
     StringBuffer b = new StringBuffer("OR(");
     Enumeration n = elems.elements();
     while (n.hasMoreElements()) {
        Element e = (Element)n.nextElement();
        b.append(e.asString());
        b.append(n.hasMoreElements() ? "|" : ")");
     }
     return b.toString();
  }
  
  /** Returns the union of all variants of all elements in the
      OrElement.
  */
  public Vector variants() {
     Vector result = new Vector();
     Enumeration n = elems.elements();
     while (n.hasMoreElements()) {
        Element e = (Element)n.nextElement();
        Vector nextpart = e.variants();
        Enumeration nn = nextpart.elements();
        result.ensureCapacity(result.size() + nextpart.size());
        while (nn.hasMoreElements())
           result.addElement(nn.nextElement());
     }
     return result;
  }
  
  /** Calls the stringAction of the ElementAction given as argument.
#M      *** DESIGN PATTERN: ***
#M      perform() is the dispatch operation for Visitor (ElementAction).
  */
  void perform(ElementAction a) {
     a.orAction(this);
  }
}


//------------------------ ElementAction ------------------------------
/**
  Interface for operations on Element structures that must handle the
  parts differently, depending on their type.
#M  *** DESIGN PATTERN: ***
#M  Design pattern Visitor: ElementAction is the superclass of all
#M  visitors of 'Element' Composite data structures.
*/
interface ElementAction {
  void stringAction(StringElement e);
  void andAction(AndElement e);
  void orAction(OrElement e);
}


//------------------------- Depth ------------------------------
/**
  Class for computing maximum depths of the different kinds of nodes in
  an And/Or Element tree.
#M  *** DESIGN PATTERN: ***
#M  Depth is a Visitor of the Composite data structure Element.
#M  Element.perform() is the double dispatch operation of the Visitor.
*/
class Depth implements ElementAction {
  public  int maxDepth      = -1;  // Depth of deepest Element
  public  int maxAndDepth   = -1;  // Depth of deepest AndElement
  public  int maxOrDepth    = -1;  // Depth of deepest OrElement
  private int depth         =  0;  // current depth (root == 0)
  
  /** Walks through Element e to initialize itself.
      Components maxDepth, maxAndDepth, and maxOrDepth can be queried
      afterwards.
  */
  public Depth(Element e) {
     e.perform(this);
  }
  
  public void stringAction(StringElement e) {
     if (maxDepth < depth)
        maxDepth = depth;
  }
  
  public void andAction(AndElement u) {
     if (maxAndDepth < depth) {
        maxAndDepth = depth;
        if (maxDepth < depth)
           maxDepth = depth;
     }
     iterate(u.elems.elements());
  }
  
  
  public void orAction(OrElement o) {
     if (maxOrDepth < depth) {
        maxOrDepth = depth;
        if (maxDepth < depth)
           maxDepth = depth;
     }
     iterate(o.elems.elements());
  }
  
  /** Handle all elements contained in an AndElement or OrElement.
  */
  private void iterate(Enumeration n) {
     depth++;  // these elements are one level deeper
     while (n.hasMoreElements()) {
        Element e = (Element)n.nextElement();
        e.perform(this);
     }
     depth--;
  }
}


//------------------------ ElementMain ------------------------------
/**
  main program. generates an And/Or sequence and prints it.
*/
public class ElementMain {
  public static void main (String args[]) {
     AndElement u = new AndElement();
     OrElement modal = new OrElement();
     OrElement verb  = new OrElement();
     u.add(new StringElement("Who "));
     modal.add(new StringElement("can "));
     modal.add(new StringElement("will "));
     u.add(modal);
     u.add(new StringElement("after all "));
     verb.add(new StringElement("think "));
     verb.add(new StringElement("act "));
     u.add(verb);
     u.add(new StringElement("?"));

     u.print();
     System.out.println(u.asString());
     Depth t = new Depth(u);
     System.out.println("  Depth=" + t.maxDepth +
                        "  AndDepth=" + t.maxAndDepth +
                        "  OrDepth=" + t.maxOrDepth);
  }
}
