/* Tuple
   #Mm1744   
   Lutz Prechelt, Barbara Unger
   1996-12-05
   RCS: $Id: TupleMain.java,v 1.2 1997/01/22 16:42:12 prechelt Exp prechelt $
*/
import java.awt.*;
import java.util.Enumeration;
import java.util.Vector;

/*
This program manages sets of tuples.
A tuple consists of several fields, e.g. name, firstname, telephone number.
There are potentially several such tuple types and for each of them 
there is also a tuple set type.
Furthermore, the program contains GUI classes for reading tuples of a 
particular type and for displaying tuple sets of a particular tuple type.
Currently only the tuple type NTTuple is fully supported.

This is the class hierarchy:
  interface Tuple
     class NTTuple
     class NATuple
  class Tupleset
  class java.awt.Frame
     class NTReader
     class NAReader
     abstract class TupleDisplay
        class NTTupleDisp1
        abstract class TupleDispA
           class NTTupleDisp2
     class TupleMain
#M
#M*** DESIGN PATTERN: ***
#M
#MTupleDisplay is the abstract Superclass of **Observer**
#Mfor the data structure Tupleset.
#M
#MTupleDispA is the abstract Superclass of a number of such observers
#Mthat differ with respect to selection, ordering, and formatting
#Mof the tuples to be shown and use a **Template Method** to vary 
#Mthese aspects.
*/


//------------------------ Tuple ------------------------------
/**
  Common interface of all tuple types.
*/
interface Tuple {
  /** Reads a tuple and delivers it to the given Tupleset.
  */
  public void getTuple(Tupleset m);
}

//------------------------ NTTuple ------------------------------
/**
  * An implementation of Tuple for names and telephone numbers.
  */
class NTTuple implements Tuple {
  private String name;      // Name einer Person
  private String firstname;   // Firstname der Person
  private String telephone;   // Telephonenumber der Person

  public String getName() { return name; }

  public String getFirstname() { return firstname; }

  public String getTelephone() { return telephone; }

  public void setName(String n) { name = n; }

  public void setFirstname(String v) { firstname = v; }
  
  public void setTelephone(String t) { telephone = t; }
  
  /** Reads a Tuple. The exact strategy used is encapsulated
      in getTuple. getTuple only guarantees, that the Tuple
      will be announced to Tupleset s after reading and
      that its fields contain the correct values afterwards.
      getTuple does NOT guarantee that reading is already
      done when getTuple returns.
  */
  public void getTuple(Tupleset s) {
     NTReader e = new NTReader(this, s); // read in with GUI, asynchronously
  }
}


//------------------------ NTReader ------------------------------
/**
    GUI reader class for NTTuple.
    Only here the final decision of using a GUI for reading is being made.
*/
final class NTReader extends Frame {
  private Button     ok;
  private TextField  t1, t2, t3;
  private Label      l1, l2, l3;
  private NTTuple    tuple;
  private Tupleset s;

  /** Generates window with three text fieldss and an OK button
      for reading the element values of an NTTuple.
  */
  public NTReader(NTTuple t, Tupleset s) {
     super("Read in Tuple");
     Panel p = new Panel();
     p.setLayout(new GridLayout(3, 2));
     this.setLayout(new FlowLayout());
     l1 = new Label("Firstname");
     l2 = new Label("Name");
     l3 = new Label("Telephone");
     t1 = new TextField();
     t2 = new TextField();
     t3 = new TextField();
     ok = new Button("O.K.");
     p.add(l1);  p.add(t1);  // enter label/text pairs into p in 3 lines
     p.add(l2);  p.add(t2);
     p.add(l3);  p.add(t3);
     this.add(p);
     this.add(ok);
     this.tuple = t; // is needed when OK button was pressed
     this.s = s;     // ditto
     this.pack();    // compute layout
     this.show();    // show dialog window
  }

  /** Event handling method. Puts values read into Tuple when OK button
      was pressed and announces Tuple to Tupleset.
  */
  public boolean action(Event ev, Object target) {
     if (ev.id == Event.ACTION_EVENT) {
       if (ev.target == ok) {      // OK button pressed
         tuple.setFirstname(t1.getText());
         tuple.setName(t2.getText());
         tuple.setTelephone(t3.getText());
         s.newTuple(tuple);
         this.hide();              // remove Frame from screen
         this.dispose();           // free its resources
         return (true);
       }
       else {
         return (false);
       }
     }
     else {
       return (false);
     }
  }
}


//------------------------ NATuple ------------------------------
/**
    An implementation of Tuple for names and addresses
*/
class NATuple implements Tuple {
  private String name;      // Name einer Person
  private String firstname;   // Firstname der Person
  private String street;   // Street und Hausnumber der Person
  private String city;       // Postleitzahl und City der Person

  public String getName() { return name; }

  public String getFirstname() { return firstname; }

  public String getStreet() { return street; }

  public String getCity() { return city; }

  public void setName(String n) { name = n; }

  public void setFirstname(String f) { firstname = f; }
  
  public void setStreet(String s) { street = s; }

  public void setCity(String c) { city = c; }
  
  /** Reads a Tuple. The exact strategy used is encapsulated
      in getTuple. getTuple only guarantees, that the Tuple
      will be announced to Tupleset s after reading and
      that its fields contain the correct values afterwards.
      getTuple does NOT guarantee that reading is already
      done when getTuple returns.
  */
  public void getTuple(Tupleset s) {
     NAReader e = new NAReader(this, s); // read in with GUI, asynchronously
  }
}


//------------------------ NAReader ------------------------------
/**
    GUI reader class for NTTuple.
    Only here the final decision of using a GUI for reading is being made.
*/
final class NAReader extends Frame {
  private Button     ok;
  private TextField  t1, t2, t3, t4;
  private Label      l1, l2, l3, l4;
  private NATuple    tuple;
  private Tupleset s;

  /** Generates window with three text fieldss and an OK button
      for reading the element values of an NTTuple.
  */
  public NAReader(NATuple t, Tupleset s) {
     super("Read in Tuple");
     Panel p = new Panel();
     p.setLayout(new GridLayout(4, 2));
     this.setLayout(new FlowLayout());
     l1 = new Label("Firstname");
     l2 = new Label("Name");
     l3 = new Label("Street");
     l4 = new Label("City");
     t1 = new TextField();
     t2 = new TextField();
     t3 = new TextField();
     t4 = new TextField();
     ok = new Button("O.K.");
     p.add(l1);  p.add(t1);  // enter label/text pairs into p
     p.add(l2);  p.add(t2);
     p.add(l3);  p.add(t3);
     p.add(l4);  p.add(t4);
     this.add(p);
     this.add(ok);
     this.tuple = t; // is needed when OK button was pressed
     this.s = s;     // ditto
     this.pack();    // compute layout
     this.show();    // show dialog window
  }

  /** Event handling method. Puts values read into Tuple when OK button
      was pressed and announces Tuple to Tupleset.
  */
  public boolean action(Event ev, Object target) {
     if (ev.id == Event.ACTION_EVENT) {
       if (ev.target == ok) {      // OK button pressed
         tuple.setFirstname(t1.getText());
         tuple.setName(t2.getText());
         tuple.setStreet(t3.getText());
         tuple.setCity(t4.getText());
         s.newTuple(tuple);
         this.hide();              // remove Frame from screen
         this.dispose();           // free its resources
         return (true);
       }
       else {
         return (false);
       }
     }
     else {
       return (false);
     }
  }
}


//------------------------ Tupleset ------------------------------
/** 
    Manages a lot of Tuples and allows to add individual Tuples to
    this lot.
    Furthermore, manages a lot of Tupledisplays, which are called
    when a Tuple is being added.
*/
class Tupleset {
  private Vector tupledisplayvector = new Vector();  // registered Displays
  private Vector tuplevector = new Vector();         // the Tuples
    
  /** called by a Tuple generator in order to bring a new Tuple
      into the Tupleset.
  */
  synchronized void newTuple(Tuple t) {
     tuplevector.addElement((Object)t);
     Enumeration e = tupledisplayvector.elements();
     while (e.hasMoreElements())
        ((TupleDisplay)e.nextElement()).newTuple(t);
  }

  void newDisplay(TupleDisplay td){
    tupledisplayvector.addElement((Object)td);
    td.newTuples(tuplevector);
  }
}


//------------------------ TupleDisplay ------------------------------
/**
    TupleDisplay shows multiple Tuples in a separate window.
    Each subclass defines its own presentation style.
    TupleDisplays guarantee that they do not modify the Tuple objects
    (so that it is possible to commit originals to the TupleDisplay's
    custody instead of creating copies).
    A program may create a number of TupleDisplay objects and give the
    same Tuple objects to each of them.
*/
abstract class TupleDisplay extends Frame {
  private Vector tuplevector = new Vector();
  
  TupleDisplay(String name) {
     super(name);
  }
     
  /** Is called by the user of the TupleDisplay to announce
      that a new Tuple shall be added to the ones already displayed.
  */
  abstract synchronized void newTuple(Tuple t);

  /** Is called by the user of the TupleDisplay to announce
      that multiple new Tuples shall be added to the ones already displayed.
      The caller guarantees that all objects in the Vector are
      Tuple objects.
      newTuples() guarantees that it does not change 'newtuples'
  */
  abstract synchronized void newTuples(Vector newtuples);
}


//------------------------ NTTupleDisp1 ------------------------------
/**
    Displays NTTuple objects in the order in which they are delivered,
    showing all their components in a simple format.
    Using Tuple objects of other Tuple types results in 
    ClassCastException.
*/
final class NTTupleDisp1 extends TupleDisplay {
  private TextArea textarea;

  /** opens new window with a textarea in it, where Tuples will be
      displayed later on.
  */
  NTTupleDisp1(String name) {
     super(name);
     this.setLayout(new FlowLayout());
     textarea = new TextArea(5, 15);
     this.add(textarea);  // textarea is the only element in Frame
     this.pack();
     this.show();
  }

  synchronized void newTuple(Tuple t) {
     NTTuple nt = (NTTuple)t;
     // just append new Tuple to current display:
     textarea.appendText(nt.getFirstname() + " " +
                         nt.getName()    + ", " +
                         nt.getTelephone() + "\n");
  }

  synchronized void newTuples(Vector newtuples) {
     Enumeration e = newtuples.elements();
     while (e.hasMoreElements())
        newTuple((Tuple)e.nextElement());
  }
}


//------------------------ TupleDispA ------------------------------
/**
    TupleDispA displays Tuplesets in a style that can be adapted
    to many different purposes by inserting an appropriate
    selection method (reject Tuples completely), 
    comparison method (define a Tuple order) and
    formatting method (select Tuple components and format them).
*/
abstract class TupleDispA extends TupleDisplay {
  /** tuplevector contains all Tuples that are to be displayed (after 
      filtering!), their order represents the presentation order.
      tuplestringvector parallels tuplevector and contains the
      final display presentation of the Tuples.
      The contents of tuplestringvector are written into textarea
      as is. textarea contains the final display contents.
  */
  private Vector tuplevector;       // Tuples to display
  private Vector tuplestringvector; // corresponding: their representation
  private TextArea textarea;

  TupleDispA(String name) {
     super(name);
     tuplevector = new Vector();
     tuplestringvector = new Vector();
     this.setLayout(new FlowLayout());
     textarea = new TextArea(4, 16);
     this.add(textarea);  // textarea is the only element in Frame
     this.pack();
     this.show();
  }

  /** Returns desired representation of Tuple a as a String.
      The String may contain zero, one, or more newline characters.
  */
  abstract String format(Tuple a);

  /** returns true, if the Tuple should be displayed and
              false otherwise
  */
  abstract boolean select(Tuple a);

  /** returns true, if Tuple a should be displayed before Tuple b and
              false, if Tuple b should be displayed before Tuple a.
  */
  abstract boolean compare(Tuple a, Tuple b);

  /** implements adding a new Tuple.
      First select() is used to test, whether the Tuple should be added
      at all, then mergeIn() moves it to the right place in the
      presentation using compare and format() converts it into a String
      of the desired display format.
#M      *** DESIGN PATTERN: ***
#M      newTuple() together with its auxiliary method mergeIn() forms a
#M      **Template Method**. The empty spots that are filled in subclasses
#M      are the methods select(), format(), and compare().
  */
  synchronized void newTuple(Tuple t) {
     Tuple nt = (Tuple)t;
     if (!select(nt))
        return;   // this Tuple shall not be displayed
     mergeIn(nt);  // puts nt into tuplevector and tuplestringvector
     display();
  }

  /** Analog to newTuple(), but uses a loop for adding several Tuples
      at once. The actual display is updated only once at the end.
  */
  synchronized void newTuples(Vector newtuples) {
     Enumeration e = newtuples.elements();
     while (e.hasMoreElements()) {
        Tuple nt = (Tuple)e.nextElement();
        if (select(nt))
           mergeIn(nt);
        else
           ; // ignore Tuple
     }
     display();
  }

  /** Finds the place in tuplevector where the new Tuple should be
      and inserts it there. Generates the final display representation
      of the Tuple and inserts that in tuplestringvector.
  */
  private void mergeIn(Tuple nt) {
    int where = 0;  // index at which new Tuple has to be inserted.
    while(where < tuplevector.size() &&
          compare((Tuple)tuplevector.elementAt(where), nt))
      where++;
    // now 'where' indicates the target index.
    // move all other elements one position towards the end in
    // both tuplevector and tuplestringvector!
    int sz = tuplevector.size(),
        k  = sz-1;  // start at end
    tuplevector.setSize(sz+1);
    tuplestringvector.setSize(sz+1);
    while (k >= where) {
       tuplevector.setElementAt(tuplevector.elementAt(k), k+1);
       tuplestringvector.setElementAt(tuplestringvector.elementAt(k), k+1);
       k--;
    }
    // now insert new Tuple into tuplevector and its formatted counterpart
    // into tuplestringvector at position 'where':
    tuplevector.setElementAt(nt, where);
    tuplestringvector.setElementAt(format(nt), where);
  }

  /** Displays the contents of tuplestringvector in textarea.
  */
  private void display() {
     StringBuffer sb = new StringBuffer();
     Enumeration e = tuplestringvector.elements();
     sb.ensureCapacity(textarea.getText().length() + 100); // optimized!!!
     while (e.hasMoreElements())
        sb.append((String)e.nextElement());
     textarea.setText(sb.toString());  // completely new text for textarea
  }
}


//------------------------ NTTupleDisp2 ------------------------------
/**
    NTTupleDisp2 displays NTTuple, where
    1. Tuples with an empty telephone number are left out and
    2. Tuples are sorted by (last)name
    Using Tuple objects of other Tuple types results in 
    ClassCastException.
#M    *** DESIGN PATTERN: ***
#M    NTTupleDisp2 completes the **Template Method** newTuple()
#M    of TupleDispA
*/
final class NTTupleDisp2 extends TupleDispA {

  NTTupleDisp2(String name) {
     super(name);
  }

  /** returns the desired representation of Tuple a as a single line String.
  */
  String format(Tuple a) {
     NTTuple b = (NTTuple)a;
     return (b.getFirstname() + " " +
             b.getName()    + ", Tel.: " +
             b.getTelephone() + "\n");
  }

  /** Tuple shall be displayed only if it contains a telephone number.
  */
  boolean select(Tuple a) {
     return !((NTTuple)a).getTelephone().equals("");
  }

  /** a shall be displayed before b, if its 'name' is alphabetically smaller.
  */
  boolean compare(Tuple a, Tuple b) {
     return (((NTTuple)a).getName().compareTo(((NTTuple)b).getName()) < 0);
  }
}


//------------------------ TupleMain ------------------------------
/**
    Main program. Generates a main window with two buttons, one
    Tupleset and two TupleDisplays. One of the buttons creates an
    NTTuple and adds it to the Tupleset.
    There is no static type safety between the actual Tuple type
    stored in the Tupleset and the Tuple type that is expected by
    the TupleDisplays.
#M    *** DESIGN PATTERN: ***
#M    The two TupleDisplays are registered as observers at the Tupleset.
*/
public final class TupleMain extends Frame {
   Button     quit = new Button("Quit");
   Button     get  = new Button("more!");
   Tupleset allTuple = new Tupleset();

   public TupleMain(String name) {
      super(name);
      this.setLayout(new FlowLayout());
      this.add(quit);
      this.add(get);
      this.pack();
      this.show();
   }   

   /** Event handling for the whole program.
   */
   public boolean action(Event ev, Object target) {
     if (ev.id == Event.ACTION_EVENT) {
       if (ev.target == quit) {  // Quit button pressed
          this.hide();       // make main window invisible
          System.exit(0);    // exit program
       }
       else if (ev.target == get) { // "more!" button pressed
          Tuple t = new NTTuple();
          t.getTuple(allTuple);
       }
       else {
         System.out.println ("ACTION_EVENT: " + ev.arg.toString()); //!!!
         return (false);
       }
     }
     else {
       System.out.println ("Event: " + ev.arg.toString() + " " + ev.id); //!!!
       return (false);
     }
     return (true);  // all but the 'return false;' cases have been handled
   }    

   /* ------------ m a i n : ------------ */
   public static void main(String args[]) {
      TupleMain mainwindow = new TupleMain("TupleLooker");
      TupleDisplay disp1 = new NTTupleDisp1("chronological");
      TupleDisplay disp2 = new NTTupleDisp2("by lastname");
      mainwindow.allTuple.newDisplay(disp1);
      mainwindow.allTuple.newDisplay(disp2);
   }
}
